"""Console script for baikalctl."""

import atexit
import json
import logging
import socket
import sys
from pathlib import Path

import click
import click.core
import uvicorn
import yaml

from .client import baikal
from .exception_handler import ExceptionHandler
from .firefox_profile import Profile
from .shell import _shell_completion
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"


def _ehandler(ctx, option, debug):
    ctx.obj = dict(ehandler=ExceptionHandler(debug))
    ctx.obj["debug"] = debug


def _cleanup():
    baikal.shutdown()


DEFAULT_URL = "http://caldav." + ".".join(socket.getfqdn().split(".")[1:]) + "/baikal"

DEFAULTS = dict(
    username="admin",
    address="0.0.0.0",
    port=8000,
    log_level="WARNING",
    url=DEFAULT_URL,
    profile_dir=Profile.DEFAULT_DIR,
    profile_create_timeout=Profile.CREATE_TIMEOUT,
)


@click.group("baikalctl", invoke_without_command=True, context_settings={"auto_envvar_prefix": "BAIKALCTL"})
@click.version_option(message=header)
@click.option("-C", "--config-file", default=Path.home() / ".baikalctl", help="config file (default ~/.baikalctl)")
@click.option("-d", "--debug", is_eager=True, is_flag=True, callback=_ehandler, help="debug mode")
@click.option("-u", "--username", help="username (default: admin)")
@click.option("-p", "--password", help="password")
@click.option("-U", "--url", help="baikal server URL")
@click.option("-a", "--api", help="api url (selects client mode)")
@click.option("-A", "--address", help="server listen address (default: 0.0.0.0)")
@click.option("-P", "--port", type=int, help="server listen port (default: 8000)")
@click.option("-l", "--log-level", default="WARNING", help="server log level (default: WARNING)")
@click.option("-v", "--verbose", is_flag=True, help="enable diagnostic output")
@click.option("--profile-create-timeout", help="profile directory")
@click.option("--profile-dir", help="profile directory")
@click.option("--cert", help="cient certificate file")
@click.option("--key", help="client certificate key file")
@click.option("--show-config", is_flag=True, help="show configuration")
@click.option(
    "--shell-completion",
    is_flag=False,
    flag_value="[auto]",
    callback=_shell_completion,
    help="configure shell completion",
)
@click.pass_context
def cli(
    ctx,
    config_file,
    username,
    password,
    url,
    api,
    address,
    port,
    profile_dir,
    cert,
    key,
    profile_create_timeout,
    log_level,
    verbose,
    debug,
    shell_completion,
    show_config,
):
    """baikalctl - admin CLI for baikal webdav/webcal server"""

    if ctx.invoked_subcommand == "version":
        click.echo(__version__)
        sys.exit(0)

    debug = ctx.obj["debug"]

    cfgdata = {}
    cfgfile = Path(config_file)
    if cfgfile.is_file():
        if verbose:
            click.echo(f"Using config file {cfgfile}", err=True)
        cfgtext = cfgfile.read_text()
        if cfgtext:
            cfgdata = yaml.safe_load(cfgfile.read_text())

    for k, v in DEFAULTS.items():
        cfgdata.setdefault(k, v)

    try:
        if not url:
            url = cfgdata["url"]
        if not username:
            username = cfgdata["username"]
        if not password:
            password = cfgdata.get("password", None)
        if not address:
            address = cfgdata["address"]
        if not port:
            port = int(cfgdata["port"])
        if not log_level:
            log_level = cfgdata["log_level"]
        if not profile_dir:
            profile_dir = cfgdata["profile_dir"]
        if not cert:
            cert = cfgdata.get("cert", None)
        if not key:
            key = cfgdata.get("key", None)
        if not profile_create_timeout:
            profile_create_timeout = cfgdata["profile_create_timeout"]

    except KeyError as ex:
        raise RuntimeError(f"Missing config value '{ex.args[0]}'")

    if show_config:
        click.echo(f"username: {username}")
        click.echo(f"password: {"'" + '*'*len(password) + "'" if password else None}")
        click.echo(f"url: {url}")
        click.echo(f"api: {api}")
        click.echo(f"address: {address}")
        click.echo(f"port: {port}")
        click.echo(f"profile_dir: {profile_dir}")
        click.echo(f"cert: {cert}")
        click.echo(f"key: {key}")
        click.echo(f"log_level: {log_level}")
        click.echo(f"verbose: {verbose}")
        click.echo(f"debug: {debug}")
        click.echo(f"log_level: {log_level}")
        click.echo(f"profile_create_timeout: {profile_create_timeout}")
        sys.exit(0)

    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help(), err=True)
        sys.exit(1)

    logging.basicConfig(level=baikal.log_level)

    baikal.log_level = log_level
    baikal.debug = debug
    baikal.verbose = verbose
    baikal.header = header

    if profile_create_timeout is not None:
        Profile.CREATE_TIMEOUT = profile_create_timeout

    if api:
        baikal.url = api
        baikal.client = True
        if cert:
            baikal.client_kwargs["cert"] = (cert, key)
    else:
        if not password:
            raise RuntimeError("Missing config value: 'password'")
        baikal.startup(url, username, password, address, port, profile_dir, cert, key)
        atexit.register(_cleanup)

    ctx.obj = baikal


@cli.command
@click.pass_obj
def users(ctx):
    """list users"""
    click.echo(json.dumps(ctx.list_users(), indent=2))


@cli.command
@click.argument("username")
@click.argument("display-name")
@click.argument("password")
@click.pass_obj
def mkuser(ctx, username, display_name, password):
    """add user account"""
    click.echo(json.dumps(ctx.add_user(username, display_name, password), indent=2))


@cli.command
@click.argument("username")
@click.pass_obj
def rmuser(ctx, username):
    """delete user account"""
    click.echo(json.dumps(ctx.delete_user(username), indent=2))


@cli.command
@click.argument("username")
@click.pass_obj
def books(ctx, username):
    """list address books for user"""
    click.echo(json.dumps(ctx.list_address_books(username), indent=2))


@cli.command
@click.argument("username")
@click.argument("name")
@click.argument("description")
@click.pass_obj
def mkbook(ctx, username, name, description):
    """add address book for user"""
    click.echo(json.dumps(ctx.add_address_book(username, name, description), indent=2))


@cli.command
@click.argument("username")
@click.argument("name")
@click.pass_obj
def rmbook(ctx, username, name):
    """delete address book for user"""
    click.echo(json.dumps(ctx.delete_address_book(username, name), indent=2))


@cli.command
@click.pass_obj
def server(ctx):
    """API server"""
    uvicorn.run(
        "baikalctl:app",
        host=ctx.address,
        port=ctx.port,
        log_level=ctx.log_level.lower(),
    )


@cli.command
@click.pass_obj
def reset(ctx):
    """restart client driver"""
    click.echo(json.dumps(ctx.reset(), indent=2))


@cli.command
@click.pass_obj
def version(ctx):
    """print version number"""
    click.echo(__version__)


@cli.command
@click.pass_obj
def status(ctx):
    """output status"""
    click.echo(json.dumps(ctx.status(), indent=2))


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



# firefox profile

import logging
import pathlib
import shlex
import subprocess
import tempfile
import time

logger = logging.getLogger(__name__)


def countFiles(dir):
    dir = pathlib.Path(dir)
    if dir.is_dir():
        return len(list(pathlib.Path(dir).glob("*")))
    return -1


def mklist(out):
    lines = []
    for line in [line.strip() for line in out.decode().split("\n")]:
        if not line:
            continue
        if line.startswith("Certificate Nickname"):
            continue
        if line.startswith("SSL,S/MIME"):
            continue
        lines.append(line)
    return [line for line in lines if line and line]


def run(cmd, **kwargs):
    kwargs.setdefault("check", True)
    return subprocess.run(shlex.split(cmd), **kwargs)


def commonName(cert):
    with tempfile.NamedTemporaryFile(suffix=".pem") as tf:
        pemCert = cert
        if pathlib.Path(cert).suffix == ".p12":
            run(f"openssl pkcs12 -in {str(cert)} -nokeys -out {tf.name} -password pass:")
            pemCert = tf.name
        cmd = f"openssl x509 -in {pemCert} -noout -subject"
        out = subprocess.check_output(shlex.split(cmd))
        out = out.decode().strip()
        if out.lower().startswith("subject=cn"):
            prefix, _, name = out.partition("=CN")
            name = name.strip(" =")
            return name
        raise RuntimeError(f"Failed to parse subject CN from certificate {cert} {repr(out)}")


class Profile:

    CREATE_TIMEOUT = 30
    STABILIZE = 2
    DEFAULT_DIR = pathlib.Path.home() / ".cache" / "baikalctl" / "profile"

    def __init__(self, dir=None, name=None):
        if name is None:
            name = "default"
        self.name = name
        if dir is None:
            dir = self.DEFAULT_DIR
        dir.mkdir(parents=True, exist_ok=True)
        self.dir = dir
        if countFiles(self.dir) < 3:
            self.create()

    def create(self):
        logger.info("Creating profile...")
        run(f"/usr/bin/firefox --headless --createprofile '{self.name} {self.dir}'")
        proc = subprocess.Popen(shlex.split(f"/usr/bin/firefox --headless --profile {self.dir} --first-startup"))
        timeout_tick = time.time() + self.CREATE_TIMEOUT
        stable = 0
        lastcount = 0
        while stable <= self.STABILIZE:
            if time.time() > timeout_tick:
                raise RuntimeError("timeout waiting for profile init")
            if proc.poll() is not None:
                raise RuntimeError("firefox exited unexpectedly")
            count = countFiles(self.dir)
            if count == lastcount:
                stable += 1
            else:
                stable = 0
            lastcount = count
            time.sleep(1)
        proc.kill()
        proc.wait()
        logger.info(f"Profile {self.name} written to {self.dir}")

    def ListCerts(self):
        certlist = mklist(subprocess.check_output(shlex.split(f"certutil -L -d sql:{str(self.dir)}")))
        certs = {}
        for certline in certlist:
            fields = certline.split()
            key = " ".join(fields[:-1])
            value = fields[-1]
            certs[key] = value
        return certs

    def AddCert(self, cert, key=None):
        certName = commonName(cert)
        logger.info("Adding client certificate...")
        with tempfile.NamedTemporaryFile(suffix=".p12") as tf:
            if pathlib.Path(cert).suffix != ".p12":
                cmd = f"openssl pkcs12 -export -in {str(cert)} -inkey {str(key)} -out {tf.name} -passout pass:"
                run(cmd)
                cert = tf.name
            cmd = f"pk12util -i {cert} -n '{certName}' -d sql:{str(self.dir)} -W ''"
            run(cmd)
        logger.info(f"Certificate {certName} added to profile {self.name}.")
        return certName



"""Top-level package for baikalctl."""

from .cli import cli
from .server import app
from .version import __author__, __email__, __timestamp__, __version__

__all__ = ["cli", "__version__", "__timestamp__", "__author__", "__email__", "app"]



import logging
from typing import Annotated

from fastapi import FastAPI, Form
from pydantic import BaseModel

from .client import baikal

app = FastAPI()

logger = logging.getLogger("uvicorn")


class User(BaseModel):
    username: str
    displayname: str
    password: str


class AddressBook(BaseModel):
    username: str
    bookname: str
    description: str


@app.on_event("startup")
def startup_event():
    logger.info(baikal.header)


@app.get("/status/")
def get_status():
    return baikal.status()


@app.post("/reset/")
def post_reset():
    return baikal.reset()


@app.get("/users/")
def get_users():
    return baikal.list_users()


@app.post("/user/")
def post_user(user: Annotated[User, Form()]):
    return baikal.add_user(user.username, user.displayname, user.password)


@app.delete("/user/{username}/")
def delete_user(username: str):
    return baikal.delete_user(username)


@app.get("/addressbooks/{username}/")
def get_addressbooks(username: str):
    return baikal.list_address_books(username)


@app.post("/addressbook/")
def post_address_book(book: Annotated[AddressBook, Form()]):
    return baikal.add_address_book(book.username, book.bookname, book.description)


@app.delete("/addressbook/{username}/{bookname}/")
def delete_addressbooks(username: str, bookname: str):
    return baikal.delete_address_book(username, bookname)



# baikalctl client

import logging
import re

import arrow
import requests
from selenium import webdriver
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.firefox_profile import FirefoxProfile

from .firefox_profile import Profile
from .version import __version__

MIN_PASSWD_LEN = 8

logger = logging.getLogger(__name__)


class Client:
    def __init__(self):
        self.driver = None
        self.logged_in = False
        self.startup_time = arrow.now()
        self.header = "baikalctl v" + __version__
        self.log_level = "WARNING"
        self.verbose = False
        self.reset_time = None
        self.profile_dir = Profile.DEFAULT_DIR
        self.client_kwargs = {}

    def startup(self, url, username, password, address, port, profile_dir, cert_file, key_file):
        logger.info("startup")
        self.client = False
        self.url = url
        self.username = username
        self.password = password
        self.address = address
        self.port = port
        self.logged_in = False
        self.profile = None
        self.profile = Profile(self.profile_dir)
        self.profile_dir = self.profile.dir
        self.cert_file = cert_file
        self.key_file = key_file
        self.client_kwargs = {}
        if cert_file is not None:
            self.profile.AddCert(cert_file, key_file)

    def _load_driver(self):
        if not self.driver:
            logger.info("load_driver")
            self.firefox_options = webdriver.FirefoxOptions()
            self.firefox_options.profile = FirefoxProfile(self.profile.dir)
            self.firefox_options.profile.set_preference("security.default_personal_cert", "Select Automatically")
            self.driver = webdriver.Firefox(options=self.firefox_options)

    def shutdown(self):
        logger.info("shutdown")
        if self.logged_in:
            self._logout()
        if self.driver:
            self.driver.quit()
            self.driver = None

    def _login(self):
        if self.logged_in:
            return
        self._load_driver()
        logger.info("login")
        try:
            self.driver.get(self.url + "/admin/")
        except WebDriverException as ex:
            return dict(error=ex.msg)
        username_text = self.driver.find_element(by=By.ID, value="login")
        password_text = self.driver.find_element(by=By.ID, value="password")
        authenticate_button = self.driver.find_element(by=By.TAG_NAME, value="button")
        if authenticate_button.text != "Authenticate":
            return dict(error="failed to find Authenticate button")
        username_text.clear()
        username_text.send_keys(self.username)
        password_text.clear()
        password_text.send_keys(self.password)
        authenticate_button.click()
        self.logged_in = True

    def _logout(self):
        if self.logged_in:
            logger.info("logout")
            self.driver.get(self.url + "/admin/")
            self.driver.find_element(by=By.LINK_TEXT, value="Logout").click()
            self.logged_in = False

    def _select_user_page(self):
        self.driver.get(self.url + "/admin/")
        users_menu = self.driver.find_element(by=By.LINK_TEXT, value="Users and resources")
        users_menu.click()

    def _parse_user_col(self, element):
        name, _, tail = element.text.partition("\n")
        display, _, email = tail.partition(" <")
        email = email.strip(">")
        return name, display, email

    def _parse_response(self, response):
        response.raise_for_status()
        return response.json()

    def list_users(self):
        logger.info("list_users")
        if self.client:
            return self._parse_response(requests.get(f"{self.url}/users/", **self.client_kwargs))
        ret = self._login()
        if ret:
            return ret
        self._select_user_page()
        users_cols = self.driver.find_elements(by=By.CLASS_NAME, value="col-username")
        ret = {}
        for col in users_cols:
            username, displayname, email = self._parse_user_col(col)
            ret[username] = dict(display_name=displayname, email=email)
        return ret

    def _set_text(self, by, value, text):
        element = self.driver.find_element(by=by, value=value)
        element.clear()
        element.send_keys(text)

    def _validate_name(self, text):
        if not re.match("^[a-zA-Z][a-zA-Z0-9@_-]*$", text):
            return dict(error=f"illegal characters in: '{text}'")
        return None

    def _validate_description(self, text):
        if not re.match("^[a-zA-Z][a-zA-Z0-9@_ -]*$", text):
            return dict(error=f"illegal characters in: '{text}'")
        return None

    def _validate_email(self, text):
        if not re.match("^[a-z][a-z0-9-]*@[a-z][a-z0-9\\.]*\\.[a-z][a-z]*$", text):
            return dict(error=f"illegal characters in: '{text}'")
        return None

    def _validate_password(self, text):
        if len(text) < MIN_PASSWD_LEN:
            return dict(error=f"password too short; minimum={MIN_PASSWD_LEN}")
        return None

    def add_user(self, username, displayname, password):
        logger.info(f"add_user {username} {displayname} ************")
        err = self._validate_email(username)
        if err:
            return err
        err = self._validate_description(displayname)
        if err:
            return err
        err = self._validate_password(password)
        if err:
            return err
        if self.client:
            return self._parse_response(
                requests.post(
                    self.url + "/user/",
                    data=dict(username=username, displayname=displayname, password=password),
                    **self.client_kwargs,
                )
            )
        self._login()
        self._select_user_page()
        add_user_button = self.driver.find_element(by=By.LINK_TEXT, value="+ Add user")
        add_user_button.click()
        self._set_text(By.NAME, "data[username]", username)
        self._set_text(By.NAME, "data[displayname]", displayname)
        self._set_text(By.NAME, "data[email]", username)
        self._set_text(By.NAME, "data[password]", password)
        self._set_text(By.NAME, "data[passwordconfirm]", password)
        save_changes_button = [
            b for b in self.driver.find_elements(By.CLASS_NAME, "btn-primary") if b.text == "Save changes"
        ][0]
        save_changes_button.click()
        return dict(added_user=username)

    def _find_user_column(self, username):
        self._login()
        self._select_user_page()
        table = self.driver.find_element(By.CLASS_NAME, "users")
        users_cols = table.find_elements(by=By.CLASS_NAME, value="col-username")
        for i, col in enumerate(users_cols):
            name, display, email = self._parse_user_col(col)
            if name == username:
                return i, table, col, None
        return -1, None, None, dict(error=f"not found: '{username}'")

    def delete_user(self, username):
        logger.info(f"delete_user {username}")
        if self.client:
            return self._parse_response(requests.delete(f"{self.url}/user/{username}/", **self.client_kwargs))
        i, table, col, err = self._find_user_column(username)
        if err:
            return err
        table.find_elements(By.CLASS_NAME, "col-actions")[i].find_element(By.LINK_TEXT, "Delete").click()
        buttons = self.driver.find_elements(By.CLASS_NAME, "btn-danger")
        for button in buttons:
            if button.text == "Delete " + username:
                button.click()
                return dict(deleted_user=username)
        return dict(error="not found: '{username}'")

    def list_address_books(self, username):
        logger.info(f"list_address_books {username}")
        if self.client:
            return self._parse_response(requests.get(f"{self.url}/addressbooks/{username}/", **self.client_kwargs))
        i, table, col, err = self._find_user_column(username)
        if err:
            return err
        table.find_elements(By.CLASS_NAME, "col-actions")[i].find_element(By.LINK_TEXT, "Address Books").click()
        names = self.driver.find_elements(By.CLASS_NAME, "col-displayname")
        contacts = self.driver.find_elements(By.CLASS_NAME, "col-contacts")
        descriptions = self.driver.find_elements(By.CLASS_NAME, "col-description")
        books = {}
        for i, name in enumerate(names):
            books[name.text] = dict(contacts=int(contacts[i].text), description=descriptions[i].text)
        return books

    def add_address_book(self, username, name, description):
        logger.info(f"add_address_book {username} {name} {description}")
        err = self._validate_email(username)
        if err:
            return err
        err = self._validate_description(name)
        if err:
            return err
        err = self._validate_description(description)
        if err:
            return err
        if self.client:
            return self._parse_response(
                requests.post(
                    self.url + "/addressbook/",
                    data=dict(username=username, bookname=name, description=description),
                    **self.client_kwargs,
                )
            )
        token = name + " " + description
        token = token.replace(" ", "-")
        i, table, col, err = self._find_user_column(username)
        if err:
            return err
        table.find_elements(By.CLASS_NAME, "col-actions")[i].find_element(By.LINK_TEXT, "Address Books").click()
        self.driver.find_element(by=By.LINK_TEXT, value="+ Add address book").click()
        self._set_text(By.NAME, "data[uri]", token)
        self._set_text(By.NAME, "data[displayname]", name)
        self._set_text(By.NAME, "data[description]", description)
        save_changes_button = [
            b for b in self.driver.find_elements(By.CLASS_NAME, "btn-primary") if b.text == "Save changes"
        ][0]
        save_changes_button.click()
        return dict(added_address_book=name)

    def delete_address_book(self, username, name):
        logger.info(f"delete_address_book {username} {name}")
        if self.client:
            return self._parse_response(
                requests.delete(f"{self.url}/addressbook/{username}/{name}/", **self.client_kwargs)
            )
        i, table, col, err = self._find_user_column(username)
        if err:
            return err
        table.find_elements(By.CLASS_NAME, "col-actions")[i].find_element(By.LINK_TEXT, "Address Books").click()
        names = self.driver.find_elements(By.CLASS_NAME, "col-displayname")
        buttons = self.driver.find_elements(By.CLASS_NAME, "btn-danger")
        for j, button in enumerate(buttons):
            if names[j].text == name:
                button.click()
                for b in self.driver.find_elements(By.CLASS_NAME, "btn-danger"):
                    if b.text == "Delete " + name:
                        b.click()
                        return dict(deleted_address_book=name)
        return dict(error=f"not found: '{name}'")

    def reset(self):
        logger.info("reset")
        if self.client:
            return self._parse_response(requests.post(f"{self.url}/reset/", **self.client_kwargs))
        self.shutdown()
        self.startup(
            baikal.url,
            baikal.username,
            baikal.password,
            baikal.address,
            baikal.port,
            baikal.profile_dir,
            baikal.cert_file,
            baikal.key_file,
        )

        self.reset_time = arrow.now()
        return dict(message="server reset")

    def status(self):
        logger.info("status")
        if self.client:
            return self._parse_response(requests.get(f"{self.url}/status/", **self.client_kwargs))
        return dict(
            name="baikalctl",
            version=__version__,
            driver=repr(self.driver),
            uptime=self.startup_time.humanize(),
            reset=self.reset_time.humanize() if self.reset_time else "never",
            profile_dir=self.profile.name if self.profile else None,
            certificates=list(self.profile.ListCerts().keys()) if self.profile else None,
            certificate_loaded=self.cert_file,
            url=self.url,
            username=self.username,
            password="*" * len(self.password),
            log_level=self.log_level,
            verbose=self.verbose,
            header=self.header,
        )


baikal = Client()



import logging

import pytest

from baikalctl.firefox_profile import Profile, countFiles, run

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")


@pytest.fixture
def keypair(shared_datadir):
    cert = shared_datadir / "test.pem"
    key = shared_datadir / "test.key"
    run(
        "openssl req -x509 -nodes -days 365 -newkey rsa:2048"
        f" -keyout {str(key)} -out {str(cert)} -subj '/CN=test_client_certificate'"
    )
    return (cert, key)


def test_profile_create(shared_datadir, keypair):
    pem, key = keypair
    assert pem.is_file()
    assert key.is_file()
    dir = shared_datadir / "profile"
    profile = Profile(dir)
    assert dir.is_dir()
    assert dir == profile.dir
    assert len(list(dir.glob("*"))) > 3
    logger.info(f"profile created with {countFiles(dir)} files")

    before = profile.ListCerts()
    for i, cert in enumerate(before):
        logger.info(f"before[{i}]: {cert}")

    profile.AddCert(pem, key)
    after = profile.ListCerts()

    for i, cert in enumerate(after):
        logger.info(f"after[{i}]: {cert}")

    assert len(after) == len(before) + 1



# cli tests

import json
import os
import shlex

import pytest
from click.testing import CliRunner

import baikalctl
from baikalctl import __version__, cli


def test_version():
    """Test reading version and module name"""
    assert baikalctl.__name__ == "baikalctl"
    assert __version__
    assert isinstance(__version__, str)


@pytest.fixture
def run():
    runner = CliRunner()

    env = os.environ.copy()
    env["TESTING"] = "1"

    def _run(cmd, **kwargs):
        assert_exit = kwargs.pop("assert_exit", 0)
        assert_exception = kwargs.pop("assert_exception", None)
        parse_json = kwargs.pop("parse_json", True)
        env.update(kwargs.pop("env", {}))
        env.update({"BAIKAL_API": "http://localhost:8000"})
        kwargs["env"] = env
        result = runner.invoke(cli, cmd, **kwargs)
        if assert_exception is not None:
            assert isinstance(result.exception, assert_exception)
        elif result.exception is not None:
            raise result.exception from result.exception
        elif assert_exit is not None:
            assert result.exit_code == assert_exit, (
                f"Unexpected {result.exit_code=} (expected {assert_exit})\n"
                f"cmd: '{shlex.join(cmd)}'\n"
                f"output: {str(result.output)}"
            )
        if parse_json:
            return json.loads(result.output)
        return result

    return _run


def test_api_client(run):

    result = run(["mkuser", "test1@example.org", "test1 user", "testpassword"])
    assert result == dict(added_user="test1@example.org")

    result = run(["users"])
    assert isinstance(result, dict)
    assert "test1@example.org" in result.keys()

    result = run(["rmuser", "test1@example.org"])
    assert isinstance(result, dict)
    assert "test1@example.org" not in result.keys()

    result = run(["mkuser", "test1@example.org", "test1 user", "testpassword"])
    assert result == dict(added_user="test1@example.org")

    result = run(["mkbook", "test1@example.org", "testbook1", "testbook one description"])
    assert result == dict(added_address_book="testbook1")

    result = run(["books", "test1@example.org"])
    assert isinstance(result, dict)
    assert "testbook1" in result.keys()

    result = run(["rmbook", "test1@example.org", "testbook1"])
    assert result == dict(deleted_address_book="testbook1")
